Program KASISK;
{
	Kasiski analysis of polyalphabetic ciphers (Vigenre,
	  Beaufort, Variant).

	Ciphertext is read from file, not keyboard.  (If anything
	  goes wrong, it's burdensome to have to renter text
	  again & again.)  See input file format below.

	Does Kasiski analysis & lists intervals between repeated
	  substrings & factors of intervals.
	Prompts user for factor to try.
	For each factor, program shows letter frequencies for each
	  column together with index of coincidence (phi) & (for
	  reference) values of phi for random & monoalphabetic data.
	  Does cross-correlation with English letter frequencies
	  to determine three most likely key letters.
	Program prompts user for acceptance & (if accepted) for
	  keyword; displays ciphertext & attempted decryption in
	  matrix form.
	References: for Kasiski, Kahn, "The Codebreakers" & Gaines,
	  "Elementary Cryptanalysis"; for index of coincidence, Kahn.

	Text is represented as array of char in order to have
	  entire text in a single data structure.

	Program organization is essentially sequential:
	  Reads ciphertext, removes nonalphabetics & converts to
	    uniform case.
	  Does Kasiski analysis & displays factors found.  For short
	    messages, repeated substrings of length 2 are accepted;
	    for longer messages, minimum acceptable substring length
	    steadily increases.
	  Evaluates any specified key length, finding letter frequen-
	    for each column.  Uses these frequencies to compute phis
	    & probable key letters for each column.  Normalizes &
	    averages phis to get confidence level.
	  Decrypts & displays text.

	Weaknesses:
	  There ought to be some kind of editing capability
	     so user can experiment with key interactively;
	     don't have good way of doing this yet.

	Input file format:
	  First line: encryption method: V for Vigenre, B for
	    Beaufort, A for vAriant.
	  Rest of file: cipher text.  Upper or lower case,
	    grouped in any convenient way, max 78 chars/line.
	    Last line must end in <cr>.

	T. Parsons	June 22, 1991
		Revised	May  29, 1993	Streamlined logic in
					  a few places, improved
					  tabulation from Evaluate,
					  & added file output proc.
                                          Protected user if method
					  omitted in first line of
					  file.
									}
uses crt;
const
   max_lgth = 1024;	{ Max allowable length of cipher text	}
   maxf = 24;		{ Largest factor considered		}
   lastint = 100;	{ Number of intervals collected		}
   phi_r = 0.0385;	{ Kappa for random characters		}
   phi_p = 0.0667;	{ Plaintext kappa			}
   maxk = maxf;		{ Longest possible key			}
type
   fstring = string[80];
   texttype = array [1..max_lgth] of char;
   freqtype = array ['a'..'z'] of integer;
var
   ciphr,                       { Ciphertext                    }
   clear: texttype;             { Clear text                    }
   cc,
   method: char;
   infname: fstring;            { Name of file containing crypt }
   fact,
   msg_lgth: integer;
   inf: text;


function OPEN_IN (var inf: text; fname: fstring): boolean; forward;
{$I open_in.inc}
function OPEN_OUT (var outf: text; var fname: fstring): boolean; forward;
{$I open_out.inc}
{$I agree.inc}


function LOCASE(c: char): char;		{ Force uniform lower case	}
   begin
   if c in ['A'..'Z'] then
      locase := chr(ord(c) + 32)
   else
      locase := c
   end;  { LoCase }


procedure READ_NEW;             { Inputs new ciphertext			}
				{ Reads lines from user, converts them	}
   var				{   to upper case, forms into screen-	}
      i,			{   sized lines				}
      k,		{ Index into ciphr array}
      col: integer;	{ Column counter	}
      lines: integer;
      cc: char;		{ Ciphertext character	}
      input: string[80];{ Line image		}

   begin
   for k := 1 to max_lgth do
      begin			{ Wipe out old text, if any	}
      ciphr[k] := ' ';
      clear[k] := ' ';
      end;
   k := 0;
   lines := 0;
   readln (inf, input);
   if (length(input) <= 2) and (upcase(input[1]) in ['V', 'B', 'A']) then
      begin
      method := upcase(input[1]);
      lines := 1
      end
   else
      method := '?';
   writeln ('Cipher text:');
   repeat				{ Until input read in   }
      if lines > 0 then			{ If method known, or	}
					{   past first line,	}
	 readln (inf, input);		{   read a line		}
      inc (lines);
      if length(input) > 0 then
	 begin
	 writeln (' ', input);		{ Display it		}
	 for i := 1 to length(input) do
	    begin
	    cc := locase(input[i]);
	    if cc in ['a'..'z'] then
	       begin			{ Squeeze out blanks	}
	       if k < max_lgth then
		  inc (k);
	       ciphr[k] := cc;
	       end
	    end  { for }
	 end  { if }
   until eof(inf);
   msg_lgth := k;
   close (inf);
   if k = 0 then
      writeln ('No characters read.')
   else if k >= max_lgth then
      begin
      writeln ('That was ', max_lgth, ' characters long or longer.',
		'  I''ve probably run out of space,');
      writeln ('  but I have the first ', max_lgth, ' characters in',
		' memory and can decrypt them if');
      writeln ('  you want.')
      end
   else
      writeln(msg_lgth,' characters.');	{ Summarize		}
   write ('Encryption scheme: ');
   case method of
	     'V': writeln ('Vigenere');
	     'B': writeln ('Beaufort');
	     'A': writeln ('Variant');
	     '?': writeln ('Unknown');
	     else begin			{ This should never happen	}
		  writeln ('Not recognized.');
                  writeln ('  You may have omitted',
				' the encryption code in Line 1.');
		  method := '?'
		  end
	     end;  { case }
   end;  { Read_New }


procedure KASISKI;	{ Kasiski analysis procedure.  Finds repeated	}
			{   substrings in cipher text & intervals be-	}
			{   tween them.  Probable period (key length)	}
			{   is factor of one or more of these intervals.}
   var
      interval: array [1..lastint] of integer;
      factors: array [2..maxf] of integer;
      count,		{ Length of repeated substring	}
      fact,		{ (Possible) factor		}
      i,
      inct,		{ Indexes interval array	}
      ist, jst,		{ Start of repetitions		}
      ic, jc,		{ Index chars through reps	}
      kmin: integer;	{ Min acceptable length		}
   begin
   writeln ('Kasiski analysis:');
   writeln ('  Interval  String        Interval  String');
   if msg_lgth > 800 then
      kmin := 5				{ Min. acceptable	}
   else if msg_lgth > 400 then		{   length depends on	}
      kmin := 4				{   message length	}
   else if msg_lgth > 200 then
      kmin := 3
   else
      kmin := 2;
   for i := 1 to 100 do			{ Clear arrays		}
      interval[i] := 0;
   for fact := 2 to maxf do
      factors[fact] := 0;
   inct := 0;
   for ist := 1 to msg_lgth do
      for jst := ist + 2 to msg_lgth do	{ Search for repeated	}
	 begin				{   substrings (done	}
	 ic := ist;			{   by brute force)	}
	 jc := jst;
	 count := 0;
	 while (jc <= msg_lgth) and (ciphr[ic] = ciphr[jc]) do
	    begin
	    inc (count);		{ Find length of	}
	    inc (ic);			{   substring		}
	    inc (jc)
	    end;
	 if count >= kmin then		{ Log intervals for	}
	    if inct < lastint then	{   non-trivial cases	}
	       begin
	       inc(inct);
	       interval[inct] := jst - ist;
	       write (interval[inct]:6, '':8);
	       for i := ist to ist + count - 1 do
	          write(ciphr[i]);
	       if odd(inct) then	{ Format output	}
	          gotoxy (25, wherey)
               else
	          writeln
               end
	 end;  { for ist, jst }
      if odd(inct) then writeln;
      for i := 1 to inct do		{ Find factors of	}
	 for fact := 2 to maxf do	{   intervals		}
	    if interval[i] mod fact = 0 then
	       inc (factors[fact]);
      writeln ('Factors, frequencies,',	{ List them		}
               ' & weighted frequencies:');
      for i := 2 to maxf do
	 write (i:3);
      writeln;
      for i := 2 to maxf do
	 write (factors[i]:3);
      writeln;
      for i := 2 to maxf do
	 write (round(sqrt(i)*factors[i]):3);
      writeln;
   end;  { Kasiski }


procedure XCOR (var freqs: freqtype; var maxc, maxc2, maxc3: char);
				{ Does cross-correlation between column	}
				{   frequencies & English letter fre-	}
				{   quencies to find most likely key	}
				{   letters.  Used by Evaluate, below.	}
   const natfrqs: freqtype =
	 (80, 16, 32, 36,123, 23, 16,
	  51, 72,  1,  5, 40, 22,	{ These are English	}
	  72, 79, 23,  2, 60, 66, 96,	{   letter frequencies	}
	  31,  9, 20,  2, 13,  1);
   var
      i, j,
      lag: integer;	{ Current offset	}
      lmax:		{ Offsets with max corr	}
	    array[1..3] of integer;
      corr: real;	{ Current correlation	}
      cmax:		{ Three largest corrs.	}
	    array[1..3] of real;
      c1, c2: char;	{ Loop indices		}
   begin
   cmax[3] := 0;
   for lag := 0 to 25 do
      begin
      corr := 0;
      for c1 := 'a' to 'z' do		{ Compute correlations	}
	 begin
	 j := ord(c1) + lag - ord('a');
	 if j > 25 then j := j - 26;
	 if method = 'B' then		{ Special handling	}
	    j := 25 - j;		{   for Beaufort	}
	 c2 := chr(j + ord('a'));
	 corr := corr + freqs[c1]*natfrqs[c2];
	 end;
      i := 3;
      while i >= 1 do			{ Update three greatest	}
	 begin
	 if corr > cmax[i] then
	    begin
	    for j := 1 to i - 1 do
	       begin			{ Shift others down	}
	       lmax[j] := lmax[j + 1];
	       cmax[j] := cmax[j + 1]
	       end;
	    cmax[i] := corr;
	    lmax[i] := lag;
	    i := 0;			{ Quick exit from loop	}
	    end;  { if }
	 dec (i)
	 end  { while }
      end;  { for lag }
   case method of			{ Find corresponding	}
	'V': for i := 1 to 3 do		{   key letters		}
		begin
		lmax[i] := 26 - lmax[i];
		if lmax[i] > 25 then lmax[i] := lmax[i] - 26;
		end;
	'B': for i := 1 to 3 do
		begin
		lmax[i] := 25 - lmax[i];
		if lmax[i] < 0 then lmax[i] := lmax[i] + 26;
		end;
	'A': begin end		{ Ok as is	}
	end;  { case }
   maxc := chr(ord('a') + lmax[3]);
   maxc2 := chr(ord('a') + lmax[2]);
   maxc3 := chr(ord('a') + lmax[1]);
   end;  { XCor }


procedure EVALUATE (period: integer);	{ Assesses this period length	}
				{ Finds letter frequencies for each	}
				{   column together with index of coin-	}
				{   cidence & most likely key letters.	}
   var
      freqs: freqtype;	{ Letter frequencies	}
      col,		{ Current column	}
      col_hgt,		{ Number of chars in col}
      i: integer;
      delta,		{ Normalized distance	}
			{   from phi_r to phi	}
      phi,		{ Index of coincidence	}
      sig,		{ Figure of merit	}
      sum: real;	{ Sum of deltas		}
      c,
      cmax,		{ Most likely keys	}
      cmax2, cmax3: char;
   begin
   writeln;				{ Write headings	}
   writeln ('Col  Frequencies', '(phi_r =':31, phi_r:7:4, '; phi_p =',
			phi_p:7:4, ')', 'Likely':8);
   write ('':4);
   for c := 'a' to 'z' do
      write (c:2);
   writeln ('    Phi    f.m.    Keys');
   sum := 0;
   for col := 1 to period do
      begin
      write (col:2, '  ');
      for c := 'a' to 'z' do
	 freqs[c] := 0;
      i := col;
      col_hgt := 0;
      while i <= msg_lgth do		{ Accumulate		}
	 begin				{   letter frequencies	}
	 inc (freqs[ciphr[i]]);
	 inc (i, period);
	 inc (col_hgt)
	 end;
      phi := 0;				{ Compute index		}
      for c := 'a' to 'z' do		{   of coincidence	}
	 phi := phi + freqs[c]*(freqs[c] - 1);
      phi := phi/(col_hgt*pred(col_hgt));
      delta := (phi - phi_r)/(phi_p - phi_r);
      sum := sum + delta;
      xcor (freqs, cmax, cmax2, cmax3);
      for c := 'a' to 'z' do		{ Show results		}
	 write (freqs[c]:2);
      writeln (phi:8:3, delta:7:3, cmax:4, cmax2:2, cmax3:2);
      end;
   sig := sum/period;
   writeln ('Average fig. of merit (near 1 => good):', sig:7:3);
   end;  { Evaluate }


procedure DECRYPT (period: integer; method: char);
   var					{ Gets keyword from user &	}
      i, j,				{   decrypts message		}
      ic, ik, ip: integer;
      keystr: string[maxk];
   begin
   repeat				{ Get & check keyword	}
      write ('Key: ');
      readln (keystr);
      if length(keystr) <> period then
	 writeln ('Huh???');
   until length(keystr) = period;
   for i := 1 to period do		{ Convert to lower case	}
      keystr[i] := locase(keystr[i]);
   for i := 1 to period do		{ Decrypt		}
      begin
      j := i;
      while j <= msg_lgth do
	 begin
	 ic := ord(ciphr[j]) - ord('a');
	 ik := ord(keystr[i]) - ord('a');
	 case method of
	     'V': ip := ic - ik;	{ These are inverses of	}
	     'B': ip := ik - ic;	{   defining equations	}
	     'A': ip := ic + ik;	{   for encryption	}
	     end;  { case }
	 if ip < 0 then ip := ip + 26	{ Keep within range	}
	 else if ip > 25 then ip := ip - 26;
	 clear[j] := chr(ip + ord('a'));
	 inc (j, period)
	 end
      end
   end;  { Decrypt }


procedure DISPLAY (period: integer;	{ Show ciphertext & cleartext	}
	var ciphr, clear: texttype);	{    as matrices		}
   var				
      i, j,
      linecount: integer;
      c: char;
   begin
   i := 0;
   linecount := 0;
   while i <= msg_lgth do
      begin				{ Ciphertext on left	}
      for j := i + 1 to i + period do
	 if j <= msg_lgth then write (ciphr[j]:2)
			  else write (' ':2);
      write ('':6);			{ Plaintext on right	}
      for j := i + 1 to i + period do
	 if j <= msg_lgth then write (clear[j]:2)
			  else write (' ':2);
      writeln;
      inc (i, period);
      inc (linecount);
      if linecount mod 22 = 0 then
         begin
         write ('<cr> for more');
         readln
         end
      end;
   end;  { Display }


procedure WRITE_OUT;
   var
      i: integer;
      outfname: fstring;
      outf: text;
   begin
   if open_out (outf, outfname) then
      begin
      for i := 1 to msg_lgth do
	 begin
	 write (outf, clear[i]);
	 if i mod 65 = 0 then
	    writeln (outf)
	 end;
      writeln (outf);
      close (outf)
      end
   end;  { Write_Out }


{  = = = = =  M a i n   P r o g r a m  = = = = = = = = = = = = = = = =  }

begin
   lowvideo;
   writeln;
   writeln ('This is Kasisk [5-21-93].');
   while open_in (inf, infname) do
      begin
      read_new;
      if method = '?' then
	 repeat
	    write ('  Select V[igenere], B[eaufort], or [v]A[riant]: ');
	    readln (cc);
	    method := upcase(cc);
	 until method in ['A', 'B', 'V'];
      if method in ['A', 'B', 'V']	{ Need this test in case	}
	    then			{   file written incorrectly	}
	 begin
	 kasiski;		{ Find probable periods	}
	 repeat
	    repeat
	       write ('Select a factor (0 terminates): ');
{$I-}
	       readln (fact);
{$I+}
	    until ioresult = 0;
	    if fact > 0 then
	       begin			{ Show results for	}
	       evaluate (fact);		{   this factor		}
	       if agree ('Ok?') then
		  begin			{ Get keyword & decrypt	}
		  decrypt (fact, method);
					{ Show results		}
		  display (fact, ciphr, clear);
		  if agree ('Ok?') then
		     begin
		     fact := 0;		{ Quick getaway		}
		     if agree ('Write sol to file?') then
			write_out
		     end
		  end
	       end  { if fact }
	 until fact = 0
	 end  { if method }
      end;  { while }
   writeln ('Done.');
end.
