       /* NEWPLAY.C                     */
       /* Playfair cryptanalysis aid    */
       /* (c) J.W.Stumpel, June 1989    */

/* Library includes */
#include <conio.h>
#include <alloc.h>
#include <stdio.h>
#include <bios.h>
#include <ctype.h>
#include <dos.h>

/* Constant definitions */
#define VIDEO 16   /* hex 10, the video interrupt        */
#define SKIP 'J'   /* letter which will not be used      */
#define BEEP 7     /* Control-G, or Bell                 */
#define BUFMAX 2000
#define TIPMAX 40
#define DOT -20
#define NORMAL; textbackground(BLUE);textcolor(WHITE);
#define INVERSE; textbackground(LIGHTGRAY);textcolor(BLACK);
#define CURSORON; cursor(6,7);
#define CURSOROFF; cursor(20,20);



/* Global variables */
char  RN[25],CN[25],NEWRN[25],NEWCN[25];
char  TOP[25], BTM[25], NEWTOP[25], NEWBTM[25];
char  LFT[25], RGT[25], NEWLFT[25], NEWRGT[25];

/* Function prototypes */
void display(void);      /* Display tr-table  */
int dofit1(char  A, char  B, char C, char D);
int dofit2(char  A, char  B, char C, char D);
int dofit3(char  A, char  B, char C, char D);
int dofit4(char  A, char  B, char C);
int dofit5(char  A, char  B, char C);
int addrm(char  A , char  B);  /* Add B as row mate to A.         */
int addcm(char  A, char  B);  /* Add B as column mate to A.      */
int addrn(char  A, char  B);  /* Add B as row neighbour to A.    */
int addcn(char  A, char  B);  /* Add B as column neighbour to A. */
int mergerows(char start, char end, char shift, char B);
int mergecols(char start, char end, char shift, char B);
void gluerows(char starta,char enda,char startb,char endb);
void gluecols(char starta,char enda,char startb,char endb);
int thereis(int row, int column); /* Does a letter at this position exist? */
char getkey(void);
void savetr(void);
void restoretr(void);
char gettop(char a);
char getleft(char a);
char getcross(char a, char b);
int fitposs (int type, int poss, char A, char B, char C, char D);
int findtip(char *tip, char *text, int length, int N);
void fittype1(void);   /* Fit square quad           */
void fittype2(void);   /* Fit horizontal quad       */
void fittype3(void);   /* Fit vertical quad         */
void fittype4(void);   /* Fit horizontal triad      */
void fittype5(void);   /* Fit vertical triad        */
void fittyper(void);   /* Fit row mates             */
void fittypec(void);   /* Fit column mates          */
void fittypeh(void);   /* Fit horizontal neighbours */
void fittypev(void);   /* Fit vertical neighbours   */
void edittr(void);     /* Manually adjust key tables */
void nofitmessage(void);
void cursor (int start, int end);
void box (int left,int top,int right,int bottom);
void decode (char *in, char *out, int length);
void disptext(char *text, int length);
void showkey(char *txt, int begin, int length);
int putquads(int depth, char *ptip, char *ctip, char *p, char *c, int L, int F);

int intip(char *tip, int x, int y);
int instring(char *name);
int hasmates(char A);  /*does this letter have a row or column mate?*/
int findreps(int reps[], char *text, int N);


/* Start of main program */
main()
{
int k, L,j,letters,length,first,notfound;
char *cbuf, *wijzer;
char filename[64];
char *pbuf, *S;
char ptip[TIPMAX];
FILE *CIPHER;
cbuf=(char *) malloc(BUFMAX);
pbuf=(char *) malloc(BUFMAX);

/*prepare boxes*/
NORMAL;
clrscr();
CURSOROFF;
box(1,1,39,16);               /*box for ciphertext*/
gotoxy(17,1);cputs("cipher");
box(41,1,79,16);              /*box for plaintext */
gotoxy(57,1);cputs("plain");


do
{  /* This loop ends with while(3) */
int endprog=1;
/*initialize tables*/
for(j=0;j<25;j++)
   CN[j]=j,RN[j]=j,TOP[j]=0,BTM[j]=0,LFT[j]=0,RGT[j]=0;

/*clear the windows*/
NORMAL;
window(3,2,38,15);clrscr();
window(43,2,78,15);clrscr();

/*Ask the user to input the filename*/
window(1,1,80,25);
gotoxy (20,17);
cputs("P L A Y F A I R    C R Y P T A N A L Y S I S");
gotoxy(7,19);
cputs ("Enter filename: ");
notfound=1;
do
{
        gotoxy (23,19);
        INVERSE;
        cputs("                                                        ");
        gotoxy (24,19);
        CURSORON;
        if (instring(filename)<0) break;
        endprog=0;
        NORMAL;
        CIPHER=fopen(filename,"r");
        if (CIPHER==NULL)
             {
             notfound=1;
             gotoxy (18,22);
             cputs ("File not found. Press ESC.");
             while (getch()!=27);
             gotoxy (18,22);
             cputs ("                          ");
             }
        else notfound=0;
}
while (notfound);
if (endprog) break;
CURSOROFF;

/* Get the ciphertext into memory */
wijzer=cbuf;letters=0;
while ((k=getc(CIPHER)) != EOF)
        {
        if (isalpha(k))
            {
            k=toupper(k);
            if (k>SKIP) k--;
            *wijzer=k-65;
            wijzer++;
            letters++;
            if (letters >= BUFMAX) break;
            }
        }
fclose(CIPHER);
*wijzer=0;

/*display ciphertext*/
window(3,2,38,15);
disptext(cbuf,letters);

/*display plaintext */
window (43,2,78,15);
decode (cbuf,pbuf,letters);
disptext(pbuf,letters);

do {  /*This loop ends with "while(1);"  */
/*Ask the user to input the tip*/
window(1,1,80,25);
gotoxy(7,19);
cputs ("Enter tip: ");
gotoxy (18,19);
INVERSE;
cputs("                                                        ");
gotoxy (19,19);
CURSORON;

/*Get the tip */
length=intip(ptip,19,19);
if (length<0) break;
CURSOROFF;
S=cbuf,L=letters;
first=0;

/*Place the tip. User may press SPACE to find another location,
  if it exists. Esc to go back to tip input. Return to accept the tip.
  Other keys have no effect. */
do { /*This loop ends with "while(2);" */
  int f=findtip(ptip,S,L,length);
  if (f<0)
     {
     if(first==0)
         {
         char k=0;
         window(1,1,80,25);
         gotoxy(10,21);
         cputs("No match found. Press ESC, then enter new tip.");
         while (k!=27) k=getch();
         gotoxy (10,21);
         cputs("                                              ");
         break;
         }
      else
         {
         S=cbuf,L=letters,first=0;
         continue;
         }
      }
  first +=f;
  window (1,1,80,25);
  gotoxy (10,21);
  printf("Match found at position %4d             ",first);
  INVERSE;
  showkey(cbuf, first, length);
  k=getch();
  while (k==0)
       {getch();
        k=getch();}

  NORMAL;
  if (k==27)
       {
       window (1,1,80,25);
       gotoxy (10,21);
       cputs("                                              ");
       showkey(cbuf,first,length);
       break;
       }
  else if(k==13)
       {
       int p;
       do
         {
         p=putquads((length)/2, ptip, cbuf+first, pbuf, cbuf, letters, 1);
         if (p==0)
            {
            window(1,1,80,25);
            gotoxy (10,21);
            cputs ("Keyword doesn't fit - contradiction found");
            if(!getch()) getch();
            }
         else if (p>0)
            {
            window(1,1,80,25);
            gotoxy (41,21);
            cprintf("%3d",p);
            }
          }
        while (p>0);
        NORMAL;
        showkey(cbuf,first,length);
        first +=2;
        S=cbuf+first,L=letters-first+2;
        }

   else if (k==32)
      {
      NORMAL;
      showkey(cbuf,first,length);
      first +=2;
      S=cbuf+first,L=letters-first+2;
      }

} while (2);
} while (1);
} while (3);

window (1,1,80,25);
textbackground(BLACK);
textcolor(LIGHTGRAY);
clrscr();
CURSORON;
return 0;
}  /*end main*/


/* FUNCTIONS */

int fitposs (int type, int poss, char A, char B, char C, char D)
{
if (type==1)
    {
    if (poss==0) return (dofit4(B,A,C));
    else if (poss==1) return (dofit5(B,A,C));
    else return(0);
    }
else if (type==2)
    {
    if (poss==0) return (dofit4(A,B,D));
    else if (poss==1) return (dofit5(A,B,D));
    else return(0);
    }
else if (type==3)
    {
    if (poss==0) return (dofit1(A,C,D,B));
    else if (poss==1) return(dofit2(A,C,B,D));
    else if (poss==2) return(dofit3(A,C,B,D));
    }
return 0;
}


void display()
{
int i,j,skip,notready=1;
window (1,1,50,25);
clrscr();
/*prepare new display-matrix*/
savetr();
while (notready)
   {
   notready=0;
   for (j=0;j<25;j++)
        {
        if ((RGT[j]==0)&&(thereis(RN[j],CN[j]+1)))
                {
                notready=1;
                for (i=0;i<25;i++) if (CN[i]>=CN[j]+1) CN[i]++;
                }
        if ((BTM[j]==0)&&(thereis(RN[j]+1,CN[j])))
                {
                notready=1;
                for (i=0;i<25;i++) if (RN[i]>=RN[j]+1) RN[i]++;
                }
         }
    }
skip=SKIP-65;
gotoxy(1,25);cputs("Unused letters:");
j=17;
for (i=0;i<25;i++)
       {
       if (hasmates(i)) gotoxy (2*CN[i]+1,RN[i]+1);
       else {
            gotoxy(j,25);
            j++;
            }
       if (i<skip) putch(i+65);
       else putch(i+66);
       }
restoretr();
}


int dofit1(char TL, char TR, char BL, char BR)
{
return (
       addcm(TL,BL)&&addcm(TR,BR)&&
       addrm(TL,TR)&&addrm(BL,BR)
       );
}


int dofit2(char A, char B, char C, char D)
{
return (addrn(A,B)&&addrm(B,C)&&addrn(C,D));
}

int dofit3(char A, char B, char C, char D)
{
return (addcn(A,B)&&addcm(B,C)&&addcn(C,D));
}

int dofit4(char A, char B, char C)
{
return (addrn(A,B)&&addrn(B,C));
}

int dofit5(char A, char B, char C)
{
return (addcn(A,B)&&addcn(B,C));
}


int addrm(char A, char B)
/*  This function will try to add B to the "row mates"
    of A, and also add A to the "row mates" of B.
    If successful, the function returns 1, otherwise zero.           */
{
int j;

/* First see if the operation is really necessary. */
if (RN[A]==RN[B]) return(1);

/* Guarantee that B has a higher row number than A. */
if (RN[A]>RN[B]) return (addrm(B,A));

/* If A and B belong to the same structure, no success: */
if (RN[B] <= RN[A]+BTM[A]) return (0);

/* Now consider 4 different cases. First: A' structure fits
   completely in B's. */
if ((TOP[A] <= TOP[B]) && (BTM[A] <= BTM[B]))
   return (mergerows(RN[A]-TOP[A],RN[A]+BTM[A],RN[B]-RN[A],B));

/* Now the case that B's structure will fit in A's structure */
else if ((TOP[B] <= TOP[A]) && (BTM[B] <= BTM[A]))
   return (mergerows(RN[B]-TOP[B],RN[B]+BTM[B],RN[A]-RN[B],A));

/* if A's structure towers above B's : */
else if (TOP[A] > TOP[B])
   {
   /* First see if the combined structure should fold back upon itself. */
   if (TOP[A]+BTM[B] > 4)
      {
      int start=RN[A]-TOP[A], end=RN[A]-5+BTM[B], down=RN[B]-RN[A]+5;
      if (!mergerows(start,end,down,B)) return(0);

      /* Adjust the "TOP" values of the remaining members of A's structure. */
      for (j=0;j<25;j++)
         if ((RN[j] <= RN[A]+BTM[A]) && (RN[j] >= RN[A]+BTM[B]-4))
            TOP[j] -= end-start+1;
      }

   /* Glue any remaining rows which tower above B's structure to
      the top of B's structure  */
  if (TOP[A] > TOP[B])
      {
      int lift=TOP[A]-TOP[B], adeep=TOP[B]+BTM[A]+1;
      /*Adjust the BTM value of the part of A's structure below which
      B will be glued: */
      for (j=0;j<25;j++)
          if ((RN[j]>=RN[A]-TOP[A])&&(RN[j]<=RN[A]-TOP[B]-1)) BTM[j] -= adeep;
      gluerows (RN[A]-TOP[A],RN[A]-TOP[B]-1,RN[B]-TOP[B],RN[B]+BTM[B]);
      /* Adjust A's TOP value, now we have lifted off some rows */
      TOP[A] -=lift;
      }

   /* Merge A's structure (now cut down to size) into B's: */
   return (mergerows(RN[A]-TOP[A],RN[A]+BTM[A],RN[B]-RN[A],B));
   }

/* if A's structure hangs below B's : */
else if (BTM[A] > BTM[B])
   {
   /* First see if the total structure should fold back upon itself. */
   if (TOP[B]+BTM[A] > 4)
      {
      int start=RN[A]+5-TOP[B], end=RN[A]+BTM[A], down=RN[B]-RN[A]-5;
      if (!mergerows(start,end,down,B)) return(0);

      /* Adjust the "BTM" values of the remaining members of A's structure. */
      for (j=0;j<25;j++)
         if ((RN[j] >= RN[A]-TOP[A]) && (RN[j] <= RN[A]-TOP[B]+4))
            BTM[j] -= end-start+1;
      }

   /* Glue any remaining rows which hang below B's structure to
      the bottom of B's structure  */
  if (BTM[A] > BTM[B])
      {
      int lift=TOP[A]+BTM[B]+1, drop=BTM[A]-BTM[B];
      /* Adjust the TOP-values of the to-be-glued part of A's structure */
      for (j=0;j<25;j++)
         if((RN[j]>RN[A]+BTM[B])&&(RN[j]<=RN[A]+BTM[A])) TOP[j]-=lift;
      /* glue A's "hanging" part below B */
      gluerows(RN[B]-TOP[B], RN[B]+BTM[B],
               RN[A]+BTM[B]+1, RN[A]+BTM[A]);
      /*Adjust A's BTM value: */
         BTM[A]-=drop;
      }
   return (mergerows(RN[A]-TOP[A],RN[A]+BTM[A],RN[B]-RN[A],B));
   }
return 0; /*NEW JWS*/
}  /*   END addrm   */








int addcm(char A, char B)
/*  This function will try to add B to the "column mates"
    of A, and also add A to the "column mates" of B.
    If successful, the function returns 1, otherwise zero.           */
{
int j;

/* First see if the operation is really necessary. */
if (CN[A]==CN[B]) return(1);

/* Guarantee that B has a higher column number than A. */
if (CN[A]>CN[B]) return (addcm(B,A));

/* If A and B belong to the same structure, no success: */
if (CN[B] <= CN[A]+RGT[A]) return (0);

/* Now consider 4 different cases. First: A's structure fits
   completely in B's. */
if ((LFT[A] <= LFT[B]) && (RGT[A] <= RGT[B]))
   return (mergecols(CN[A]-LFT[A],CN[A]+RGT[A],CN[B]-CN[A],B));

/* Now the case that B's structure will fit in A's structure */
else if ((LFT[B] <= LFT[A]) && (RGT[B] <= RGT[A]))
   return (mergecols(CN[B]-LFT[B],CN[B]+RGT[B],CN[A]-CN[B],A));

/* if A's structure protrudes to the left of B's : */
else if (LFT[A] > LFT[B])
   {
   /* First see if the total structure should fold back upon itself. */
   if (LFT[A]+RGT[B] > 4)
      {
      int start=CN[A]-LFT[A], end=CN[A]-5+RGT[B], down=CN[B]-CN[A]+5;
      if (!mergecols(start,end,down,B)) return(0);

      /* Adjust the "LFT" values of the remaining members of A's structure. */
      for (j=0;j<25;j++)
         if ((CN[j] <= CN[A]+RGT[A]) && (CN[j] >= CN[A]+RGT[B]-4))
            LFT[j] -= end-start+1;
      }

   /* Glue any remaining columns which protrude left of B's structure to
      the left of B's structure  */
  if (LFT[A] > LFT[B])
      {
      int lift =LFT[A]-LFT[B], adeep=LFT[B]+RGT[A]+1;
      /*Adjust the RGT value of the part of A's structure to the right
      of which B will be glued: */
      for (j=0;j<25;j++)
          if ((CN[j]>=CN[A]-LFT[A])&&(CN[j]<=CN[A]-LFT[B]-1)) RGT[j] -= adeep;
      gluecols (CN[A]-LFT[A], CN[A]-LFT[B]-1,CN[B]-LFT[B], CN[B]+RGT[B]);
      /*Adjust*/
      LFT[A] -=lift;
      }

   /* Merge A's structure (now cut down to size) into B's: */
   return (mergecols(CN[A]-LFT[A],CN[A]+RGT[A],CN[B]-CN[A],B));
   }

/* if A's structure protrudes to the right of B's : */
else if (RGT[A] > RGT[B])
   {
   /* First see if the total structure should fold back upon itself. */
   if (LFT[B]+RGT[A] > 4)
      {
      int start=CN[A]+5-LFT[B], end=CN[A]+RGT[A], down=CN[B]-CN[A]-5;
      if (!mergecols(start,end,down,B)) return(0);

      /* Adjust the "RGT" values of the remaining members of A's structure. */
      for (j=0;j<25;j++)
         if ((CN[j] >= CN[A]-LFT[A]) && (CN[j] <= CN[A]-LFT[B]+4))
            RGT[j] -= end-start+1;
      }

   /* Glue any remaining columns which extend beyond the right end of B's
      structure to the right limit of B's structure  */
  if (RGT[A] > RGT[B])
      {

      int lift=LFT[A]+RGT[B]+1, drop=RGT[A]-RGT[B];
      /* Adjust the LFT-values of the to-be-glued part of A's structure */
      for (j=0;j<25;j++)
        if ((CN[j]>CN[A]+RGT[B])&&(CN[j]<=CN[A]+RGT[A])) LFT[j]-=lift;
      /* glue A's extending part to the right of B */
      gluecols(CN[B]-LFT[B], CN[B]+RGT[B],
               CN[A]+RGT[B]+1, CN[A]+RGT[A]);
      /*Adjust A's RGT value */
      RGT[A]-=drop;
      }
   return (mergecols(CN[A]-LFT[A],CN[A]+RGT[A],CN[B]-CN[A],B));
   }
return 0; /*NEW JWS*/
}  /*   END addcm  */


void gluerows(char starta,char enda,char startb,char endb)
/* glues a group of rows, beginning with row startb and ending with
   endb, below a group beginning with starta and ending with enda. */
{
int j, shift=enda-startb+1, drop=endb-startb+1, adeep= enda-starta+1;
if (enda<startb)
     {
     for (j=0;j<25;j++)
       {
       if ((RN[j]>enda)&&(RN[j]<startb)) RN[j] += drop;
       else if ((RN[j]>=startb)&&(RN[j]<=endb))
            {
            RN[j]  += shift;
            TOP[j] += adeep;
            }
       else if ((RN[j]>=starta)&&(RN[j]<=enda)) BTM[j] +=drop;
       }
     }
   else if (enda>startb)
     {
     for (j=0;j<25;j++)
       {
       if ((RN[j]>endb)&&(RN[j]<starta)) RN[j] -= drop;
       else if ((RN[j]>=starta)&&(RN[j]<=enda))
            {
            RN[j]  -= drop;
            BTM[j] += drop;
            }
       else if ((RN[j]>=startb)&&(RN[j]<=endb))
            {
            RN[j]  += shift-drop;
            TOP[j] += adeep;
            }
       }
     }
}




void gluecols(char starta,char enda,char startb,char endb)
/* glues a group of columns, beginning with col startb and ending with endb,
   to the right of a group beginning with starta and ending with enda. */
{
int j, shift=enda-startb+1, drop=endb-startb+1, adeep= enda-starta+1;
if (enda<startb)
     {
     for (j=0;j<25;j++)
       {
       if ((CN[j]>enda)&&(CN[j]<startb)) CN[j] += drop;
       else if ((CN[j]>=startb)&&(CN[j]<=endb))
            {
            CN[j]  += shift;
            LFT[j] += adeep;
            }
       else if ((CN[j]>=starta)&&(CN[j]<=enda)) RGT[j] +=drop;
       }
     }
   else if (enda>startb)
     {
     for (j=0;j<25;j++)
       {
       if ((CN[j]>endb)&&(CN[j]<starta)) CN[j] -= drop;
       else if ((CN[j]>=starta)&&(CN[j]<=enda))
            {
            CN[j]  -= drop;
            RGT[j] += drop;
            }
       else if ((CN[j]>=startb)&&(CN[j]<=endb))
            {
            CN[j]  += shift-drop;
            LFT[j] += adeep;
            }
       }
     }
}






int mergerows (char start, char end, char shift, char B)
/* Moves a block of rows from "start" to "end" downward by an amount
   "shift" and merges it with the letters already there. Return 0 if
   no room. TOP and BTM are adjusted relative to letter B. */
{
int j, count[25], drop = end-start+1;

for (j=0;j<25;j++)
   {
   count[j]=0;
   if ((RN[j] >= start) && (RN[j] <= end))
      {
      if (thereis(RN[j]+shift,CN[j])) return (0);  /* no room */
      else {
           RN[j] += shift;
           TOP[j] = TOP[B] - RN[B] + RN[j];
           BTM[j] = BTM[B] - RN[j] + RN[B];
           }
      }
   }
/* Beautify: drop A's old rows (now empty), and check no rows too long. */
for (j=0;j<25;j++)
   {
   if (RN[j] >= start) RN[j] -= drop;
   if ((++count[RN[j]])>5) return (0);  /* Overcrowded row */
   }
/*success*/
return (1);
}


int mergecols (char start, char end, char shift, char B)
/* Moves a block of columns from "start" to "end" rightward by an amount
   "shift" and merges it with the letters already there. Return 0 if
   no room. LFT and RGT are adjusted relative to letter B. */
{
int j, count[25], drop = end-start+1;
for (j=0;j<25;j++)
   {
   count[j]=0;
   if ((CN[j] >= start) && (CN[j] <= end))
      {
      if (thereis(RN[j],CN[j]+shift)) return (0);  /* no room */
      else {
           CN[j] += shift;
           LFT[j] = LFT[B] - CN[B] + CN[j];
           RGT[j] = RGT[B] - CN[j] + CN[B];
           }
      }
   }
/* Beautify: drop A's old columns (now empty), and check no columns too long. 
*/
for (j=0;j<25;j++)
   {
   if (CN[j] >= start) CN[j] -= drop;
   if ((++count[CN[j]])>5) return (0);  /* Overcrowded column */
   }
/*success*/
return (1);
}



int addrn(char A, char B)
/* Makes B the right neighbour of A */
{
int j;
/* if they are already neighbours, do nothing */
if ((RN[A]==RN[B])&&(
((CN[B]==CN[A]+1)&&(RGT[A] >= 1)) ||
((CN[B]==CN[A]-4)&&(LFT[A] == 4)))) return (1);

/* if they are in the same horizontal structure (but not neighbours)
   it it impossible for them to become neighbours: */
if ((CN[B]<=CN[A]+RGT[A])&&(CN[B] >=CN[A]-LFT[A])&&
    !((CN[B]==CN[A]+1)||(CN[A]==CN[B]+4))) return (0);

/* To be row neighbours they first have to be row mates: */
if (!addrm(A,B)) return(0);

/* see if there is a letter in the target column. */
if (RGT[A]>=1)
  {
  j=0;
  while (CN[j]!=CN[A]+1) j++;
  return (addcm(j,B));
  }

/* see if B has a protrusion to the left. */
if (LFT[B]>=1)
  {
  j=0;
  while (CN[j]!=CN[B]-1) j++;
  return (addcm(j,A));
  }

/* So A has a "smooth right" and B a "smooth left". We must now glue
   the two structures: */
    {
    int startb=CN[B], endb=CN[B]+RGT[B],
    starta=CN[A]-LFT[A], enda=CN[A],totwide;
    gluecols (starta,enda,startb,endb);
    /* If new glued structure too wide, fold back: */
    totwide=LFT[A]+RGT[A]+1;
    if (totwide>5)
        {
          int start=CN[A]-LFT[A]+5, end=CN[A]+RGT[A];
          for (j=0;j<25;j++)
          if ((CN[j]>=CN[A]-LFT[A])&&(CN[j]<start)) RGT[j] -= totwide-5;
          return(mergecols(start,end,-5,A));
        }
    else return(1);
    }

}




int addcn(char A, char B)
/* Makes B the bottom neighbour of A */
{
int j;
/* if they are already neighbours, do nothing */
if ((CN[A]==CN[B])&&(
((RN[B]==RN[A]+1)&&(BTM[A] >= 1)) ||
((RN[B]==RN[A]-4)&&(TOP[A] == 4)))) return (1);

/* if they are in the same vertical structure (but not neighbours)
   it it impossible for them to become neighbours: */
if ((RN[B]<=RN[A]+BTM[A])&&(RN[B] >=RN[A]-TOP[A])&&
    !((RN[B]==RN[A]+1)||(RN[A]==RN[B]+4))) return (0);

/* To be column neighbours they first have to be column mates: */
if (!addcm(A,B)) return(0);

/* see if there is a letter in the target row. */
if (BTM[A]>=1)
  {
  j=0;
  while (RN[j]!=RN[A]+1) j++;
  return (addrm(j,B));
  }

/* see if B has a protrusion to the top. */
if (TOP[B]>=1)
  {
  j=0;
  while (RN[j]!=RN[B]-1) j++;
  return (addrm(j,A));
  }

/* So A has a "smooth bottom" and B a "smooth top". We must now glue
   the two structures. */

    {
    int startb=RN[B], endb=RN[B]+BTM[B],
    starta=RN[A]-TOP[A], enda=RN[A],totdeep;
    gluerows (starta,enda,startb,endb);
    /* If new glued structure too high, fold back: */
    totdeep=TOP[A]+BTM[A]+1;
    if (totdeep>5)
        {
          int start=RN[A]-TOP[A]+5, end=RN[A]+BTM[A];
          for (j=0;j<25;j++)
          if ((RN[j]>=RN[A]-TOP[A])&&(RN[j]<start)) BTM[j] -= totdeep-5;
          return(mergerows(start,end,-5,A));
        }
    else return(1);
    }
}



int thereis (int row, int column)
{
int j;
for (j=0;j<25;j++) if ((RN[j]==row)&&(CN[j]==column)) return (1);
return (0);
}



char getkey()
/* Obtains letters to be used in a key quad (or triad). Only alphabetic
   byteacters, with the exception of the "skip" letter, are valid.
   A valid input will be echoed to the screen in uppercase.
   A small integer (byte) in the range of 0-24 will be returned in case
   of valid input, and -1 if ESC is pressed. */
{char k;
 CURSORON;
 k=toupper(getch());
 CURSOROFF;
 if (k==27) return(-1);
 if(isalpha(k)&&(k!=SKIP))
     {
     putch(k);
     if (k>SKIP) k--;
     return(k-65);
     }
 if (k==0) getch(); /* ALT and arrow elimination */
 return(getkey()); /* Try again.     */
 }


void savetr()
{
int i;
for (i=0;i<25;i++)
    {
     NEWRN[i]=RN[i];
     NEWCN[i]=CN[i];
     NEWTOP[i]=TOP[i];
     NEWBTM[i]=BTM[i];
     NEWLFT[i]=LFT[i];
     NEWRGT[i]=RGT[i];
    }
}

void restoretr()
{
int i;
for (i=0;i<25;i++)
     {
      RN[i]=NEWRN[i];
      CN[i]=NEWCN[i];
      TOP[i]=NEWTOP[i];
      BTM[i]=NEWBTM[i];
      LFT[i]=NEWLFT[i];
      RGT[i]=NEWRGT[i];
     }
}


void nofitmessage()
{
        textbackground (LIGHTGRAY);
        textcolor (BLACK+BLINK);
        gotoxy (11,19);
        cputs("NO FIT");
        NORMAL;
        putchar (BEEP);
        gotoxy (10,21); cputs ("Press ESC");
        while (getch()!=27);
}


void cursor (int start, int end)
{
union REGS regs;
regs.h.ah=1;
regs.h.ch=start;
regs.h.cl=end;
int86(VIDEO,&regs,&regs);
}


void box (int left,int top,int right,int bottom)
/*Draw box. Box around whole screen is box (1,1,80,25).
  Warning. Use this routine only with "correct" parameter values.
  No checking is performed. Text window must allow placement of box.  */
{int j;
gotoxy(left,top);
putch('');
for (j=1;j<(right-left);j++) putch('');
putch('');
for (j=1;j<(bottom-top);j++) {gotoxy (left,top+j);
                              putch ('');
                              gotoxy (right,top+j);
                              putch ('');}
gotoxy(left,bottom);
putch('');
for (j=1;j<(right-left);j++) putch('');
putch('');
}


void decode (char *in, char *out, int length)
{
char c0,c1,p0,p1;
int n=0;
while (n<length)
         {
         c0=*(in+n);
         c1=*(in+n+1);
              if (CN[c0]==CN[c1])    /* letters in same column */
                   {
                   p0=gettop(c0);
                   p1=gettop(c1);
                   }
               else if (RN[c0]==RN[c1])  /* letters in same row */
                   {
                   p0=getleft(c0);
                   p1=getleft(c1);
                   }
               else               /* letters in corners of rectangle */
                   {
                   p0=getcross(c0,c1);
                   p1=getcross(c1,c0);
                   }
         *(out+n)=p0;
         *(out+n+1)=p1;
         n +=2;
         }
}

char gettop (char A)
/* finds the letter above A */
{
int j=0;
while (j<25)
    {
    if ((CN[j]==CN[A])&&(((BTM[j]>=1)&&(RN[j]==RN[A]-1))||
                        (TOP[j]==4)&&(RN[j]==RN[A]+4))) return (j);
    j++;
    }
return (DOT);
}


char getleft (char A)
/* finds the letter left of A */
{
char j=0;
while (j<25)
    {
    if ((RN[j]==RN[A])&&(((RGT[j]>=1)&&(CN[j]==CN[A]-1))||
                        (LFT[j]==4)&&(CN[j]==CN[A]+4))) return (j);
    j++;
    }
return (DOT); /* space, I hope. */
}


char getcross(char A, char B)
/* finds a byteacter in A's row and B's column */
{
int j=0;
while (j<25)
    {
    if ((RN[j]==RN[A])&&(CN[j]==CN[B])) return (j);
    j++;
    }
return (DOT);
}


void disptext(char *text, int length)
/* displays a range of memory as digram text. The calling routine
   must set the window first. */
{
int n=0, skip=SKIP-65;
char k;
while (n<length)
     {
     k=*(text+n);
     if (k<skip) putch(k+65); else putch(k+66);
     n++;
     k=*(text+n);
     if (k<skip) putch(k+65); else putch(k+66);
     n++;
     putch(32);
     }
}

int intip(char *tip, int x, int y)
/*Inputs a string, stores it, displays it as digrams.
  Returns the number of valid byteacters in the string.
  Returns -1 if the user wants to abort. */
{
char k;
int j=0,index=0;
gotoxy (x,y);
INVERSE;
cputs("                                                           ");
gotoxy (x,y);
CURSORON;
do
  {
  k=getch();
  if (k==27) return(-1); /*leave*/
  else if ((k==8)&(index>0)) /*backspace pressed*/
          {
          index--;
          if(j==0) {
                    putch(8);
                    putch(8);
                    putch(32);
                    putch(8);
                    j=1;
                   }
          else     {
                   putch(8);
                   putch(32);
                   putch(8);
                   j=0;
                   }
          }
  else if (k==0) getch(); /* ALT and arrow elimination */
  else if ((isalpha(k))&&(k!=SKIP)&&(k!=SKIP+32)&&(index<TIPMAX))
     {
     k=toupper(k);
     putch (k);
     if (k>SKIP) k--;
     *(tip+index)=k-65;
     j++; if (j>1) {j=0; putch(32);}
     index++;
     }
  }
while (k!=13);
NORMAL;
return (index);
}


int findtip(char *tip, char *text, int length, int N)
/* Finds where the tip is in the ctext by matching the repeat pattern. */
{
int firstrep,ok,tipreps[TIPMAX/2],textreps[TIPMAX/2],j;
int start,found,dist;
char A,B,C,D;
firstrep=findreps(tipreps,tip,N);
if (firstrep<0)  /* No repeats. Tip may be in many places. Restrictions:
                    plain letters must always be different from cipher
                    letters; if no repeats in tip, then also no
                    repeats in ciphertext. */
     {
     start=0;
     while (start<length-N)
       {
       j=0;
       found=1;
       while (j<N)
         {
         A=*(text+start+j), B=*(tip+j);
         if (A==B)
               {
               found=0;
               break;
               }
         j++;
         }
      if ((found) && (findreps(textreps,text+start,N)<0)) return (start);
      start +=2;
      }
     return(-1); /*no match*/
     }
else {
     start=2*firstrep;
     dist=2*tipreps[firstrep];
     while (start<2*firstrep+length-N)
        {
        A=*(text+start), B=*(text+start+1);
        if (dist>0) C=*(text+start+dist), D=*(text+start+dist+1);
        else C=*(text+start-dist+1), D=*(text+start-dist);
        if ((A==C)&&(B==D))
             {
             /* We have a possible match. Now check if it is a complete
                match. */
             findreps(textreps,text+start-2*firstrep,N);
             ok=1;
             for (j=0;j<TIPMAX/2;j++)
                if(textreps[j]!=tipreps[j]) ok=0;
             if(ok) return(start-2*firstrep);
             }
        start +=2;
        }
     }
return (-1);
}

void showkey(char *txt, int begin, int length)
{
int i,j,x,y;
char k;
x=1+3*(begin%24)/2;
y=1+begin/24;
window(3,2,38,15);
gotoxy(x,y);
i=0;
for (j=begin;j<begin+length;j++)
   {
   k=*(txt+j)+65;
   if (k>=SKIP) k++;
   putch(k);
   i++;
   if ((i==2)&&(j<begin+length-1))
       {
       i=0;
       putch(32);
       }
   }
}


int putquads(int depth, char *ptip, char *ctip, char *p, char *c, int L, int F)
/* This recursive function has to be called from the outside with the
   last parameter F (for "first") set to 1.  The function returns with
   -1 (user wants to stop), or the number of valid combinations found.  */
{
char k,A,B,C,D;
char row[25],col[25],t[25],b[25],l[25],r[25];
int i,j,type,numposs;
static int excode;  /*Exit code. Final return value of the function. */
if (F) excode=0;
if (excode>=0)
  {
  /* Depth is the number of quads that we still have to fetch. */
  if (depth==0)
     {
     excode++;
     do
      {
      window(1,1,80,25);
      gotoxy (10,21);
      cprintf("Combination nr%3d found. Total:",excode);
      decode (c,p,L);
      window (43,2,78,15);
      disptext(p,L);
      k=getch();
      if (k==27)
            {
            excode=-1;
            return(-1);
            }
      if(k==13) edittr();
      }
     while (k!=32);
    }
  else
     {
     /*get a quad*/
     A=*(ptip+2*(depth-1)), B=*(ptip+2*(depth-1)+1);
     C=*(ctip+2*(depth-1)), D=*(ctip+2*(depth-1)+1);

     /*determine its type*/
     if (A==D) type=1, numposs=2;
     else if (B==C) type=2, numposs=2;
     else type=3, numposs=3;

    /*save a local copy of the key tables (on the stack) */
    for (j=0;j<25;j++)
       row[j]=RN[j], col[j]= CN[j], t[j]=TOP[j],
       b[j]=BTM[j], l[j]=LFT[j], r[j]=RGT[j];

    /*fit the quad for all its possible orientations*/
    for (i=0; i<numposs; i++)
       {
       /* the secondary incarnations must be called with F==0 */
       if (fitposs(type,i,A,B,C,D))
           putquads(depth-1, ptip, ctip,p,c,L,0);
       /*restore tables*/
       for (j=0;j<25;j++)
              RN[j]=row[j], CN[j]=col[j], TOP[j]=t[j],
              BTM[j]=b[j], LFT[j]=l[j], RGT[j]=r[j];
       }
    }
  }
return(excode);
}

void edittr()
{
char k;
char scrbuf[80*25*2];
/*save old screen*/
gettext(1,1,80,25,scrbuf);
CURSOROFF;
NORMAL;

/*clear tr-table window */
window (1,1,50,25);
clrscr();

do {
   window (51,1,80,25);
   clrscr();
   gotoxy (5,2);
   puts("KEYSQUARE EDITING");
   gotoxy (5,3);
   puts ("Action: 1-5, R, C, H, V");
   display();
   window (51,1,80,25);
   k=getch();
   while (k==0) {getch(); k=getch();}
   clrscr();
       switch (k)
        {
       case '1':  fittype1();
                  break;
       case '2':  fittype2();
                  break;
       case '3':  fittype3();
                  break;
       case '4':  fittype4();
                  break;
       case '5':  fittype5();
                  break;
       case 'r':
       case 'R':  fittyper();
                  break;
       case 'c':
       case 'C':  fittypec();
                  break;
       case 'h':
       case 'H':  fittypeh();
                  break;
       case 'v':
       case 'V':  fittypev();
                  break;
       case 27 :  break;
       default :  putchar(BEEP);
        }
   }
   while (k!=27);
window (1,1,80,25);
NORMAL;
puttext(1,1,80,25,scrbuf);
}


void fittype1()
{
char TL,TR,BL,BR;
gotoxy (7,3); cputs ("SQUARE MATCHING");
gotoxy (10,5); cputs("Ŀ");
gotoxy (10,6); cputs("      ");
gotoxy (10,7); cputs("");
gotoxy (10,8); cputs("      ");
gotoxy (10,9); cputs("");
gotoxy (3,11); cputs ("Enter the 4 byteacters");
gotoxy (3,12); cputs ("of this poss....");
gotoxy (12,6); TL=getkey(); if (TL<0) return;
gotoxy (16,6); TR=getkey(); if (TR<0) return;
gotoxy (12,8); BL=getkey(); if (BL<0) return;
gotoxy (16,8); BR=getkey(); if (BR<0) return;
savetr();
if (!dofit1(TL,TR,BL,BR))
       {
        nofitmessage();
        restoretr();
       }
}

void fittype2()
{
char A,B,C,D;
gotoxy (8,3); cputs ("HORIZONTAL QUAD");
gotoxy  (7,7); cputs("Ŀ");
gotoxy  (7,8); cputs("            ");
gotoxy  (7,9); cputs("");
gotoxy (3,11); cputs ("Enter the 4 byteacters");
gotoxy (3,12); cputs ("of this poss....");
gotoxy  (9,8); A=getkey(); if (A<0) return;
gotoxy (13,8); B=getkey(); if (B<0) return;
gotoxy (17,8); C=getkey(); if (C<0) return;
gotoxy (21,8); D=getkey(); if (D<0) return;
savetr();
if (!dofit2(A,B,C,D))
       {
        nofitmessage();
        restoretr();
       }

}



void fittype3()
/* fit a vertical quad */
{
char A,B,C,D;
gotoxy (8,3); cputs ("VERTICAL QUAD");
gotoxy  (12,4); cputs("Ŀ");
gotoxy  (12,5); cputs("   ");
gotoxy  (12,6); cputs("Ĵ");
gotoxy  (12,7); cputs("   ");
gotoxy  (12,8); cputs("");
gotoxy  (12,9); cputs("   ");
gotoxy (12,10); cputs("Ĵ");
gotoxy (12,11); cputs("   ");
gotoxy (12,12); cputs("");
gotoxy (3,13); cputs ("Enter the 4 byteacters");
gotoxy (3,14); cputs ("of this poss....");
gotoxy  (14,5); A=getkey(); if (A<0) return;
gotoxy  (14,7); B=getkey(); if (B<0) return;
gotoxy  (14,9); C=getkey(); if (C<0) return;
gotoxy (14,11); D=getkey(); if (D<0) return;
savetr();
if (!dofit3(A,B,C,D))
       {
        nofitmessage();
        restoretr();
       }

}


void fittype4()
{
char A,B,C;
gotoxy  (5,3); cputs("HORIZONTAL TRIAD");
gotoxy  (7,7); cputs("Ŀ");
gotoxy  (7,8); cputs("         ");
gotoxy  (7,9); cputs("");
gotoxy (3,11); cputs ("Enter the 3 byteacters");
gotoxy (3,12); cputs ("of this poss....");
gotoxy  (9,8); A=getkey(); if (A<0) return;
gotoxy (13,8); B=getkey(); if (B<0) return;
gotoxy (17,8); C=getkey(); if (C<0) return;
savetr();
if (!dofit4(A,B,C))
       {
        nofitmessage();
        restoretr();
       }
}



void fittype5()
{
char A,B,C;
gotoxy  (7,3);  cputs ("VERTICAL TRIAD");
gotoxy  (12,4); cputs("Ŀ");
gotoxy  (12,5); cputs("   ");
gotoxy  (12,6); cputs("Ĵ");
gotoxy  (12,7); cputs("   ");
gotoxy  (12,8); cputs("Ĵ");
gotoxy  (12,9); cputs("   ");
gotoxy (12,10); cputs("");
gotoxy (3,11); cputs ("Enter the 3 byteacters");
gotoxy (3,12); cputs ("of this poss....");
gotoxy (14,5); A=getkey(); if (A<0) return;
gotoxy (14,7); B=getkey(); if (B<0) return;
gotoxy (14,9); C=getkey(); if (C<0) return;
savetr();
if (!dofit5(A,B,C))
       {
        nofitmessage();
        restoretr();
       }
}


void fittyper()
/* Fit row mates */
{
char A,B;
gotoxy (9,3); cputs ("ROW MATES");
gotoxy  (9,7); cputs("Ŀ");
gotoxy  (9,8); cputs("      ");
gotoxy  (9,9); cputs("");
gotoxy (3,11); cputs ("Enter the 2 byteacters");
gotoxy (3,12); cputs ("of this pair....");
gotoxy (11,8); A=getkey(); if (A<0) return;
gotoxy (15,8); B=getkey(); if (B<0) return;
savetr();
if (!addrm(A,B))
       {
        nofitmessage();
        restoretr();
       }

}



void fittypec()
/* fit a pair of column mates */
{
char A,B;
gotoxy (9,3); cputs ("COLUMN MATES");
gotoxy  (12,6); cputs("Ŀ");
gotoxy  (12,7); cputs("   ");
gotoxy  (12,8); cputs("");
gotoxy  (12,9); cputs("   ");
gotoxy (12,10); cputs("");
gotoxy (3,13); cputs ("Enter the 2 byteacters");
gotoxy (3,14); cputs ("of this pair....");
gotoxy  (14,7); A=getkey(); if (A<0) return;
gotoxy  (14,9); B=getkey(); if (B<0) return;
savetr();
if (!addcm(A,B))
       {
        nofitmessage();
        restoretr();
       }

}



void fittypeh()
/* Fit row neigbours */
{
char A,B;
gotoxy (6,3); cputs ("ROW NEIGHBOURS");
gotoxy  (9,7); cputs("Ŀ");
gotoxy  (9,8); cputs("      ");
gotoxy  (9,9); cputs("");
gotoxy (3,11); cputs ("Enter the 2 byteacters");
gotoxy (3,12); cputs ("of this connected pair...");
gotoxy (11,8); A=getkey(); if (A<0) return;
gotoxy (15,8); B=getkey(); if (B<0) return;
savetr();
if (!addrn(A,B))
       {
        nofitmessage();
        restoretr();
       }
}



void fittypev()
/* fit a pair of column neighbours */
{
char A,B;
gotoxy (5,3); cputs ("COLUMN NEIGHBOURS");
gotoxy  (12,6); cputs("Ŀ");
gotoxy  (12,7); cputs("   ");
gotoxy  (12,8); cputs("Ĵ");
gotoxy  (12,9); cputs("   ");
gotoxy (12,10); cputs("");
gotoxy (3,13); cputs ("Enter the 2 byteacters");
gotoxy (3,14); cputs ("of this connected pair...");
gotoxy  (14,7); A=getkey(); if (A<0) return;
gotoxy  (14,9); B=getkey(); if (B<0) return;
savetr();
if (!addcn(A,B))
       {
        nofitmessage();
        restoretr();
       }

}

int instring (char *name)
/* Gets a string, delimited by zero. Editing by backspace only.
   Returns number of bytes in string.
   Pressing ESC aborts; string becomes null string; return value
   becomes -1. */
{int j=0; char k; char *oldname=name;
 while ((k=getch())!=13)
  {if (isprint(k)) {*name=k; putch(k); j++; name++;}
   else if ((k==8)&&(j>0)) {putch(8); putch(32); putch(8); name--; j--;}
   else if (k==27) {*oldname=0; return(-1);}
   else if (k==0) getch(); /*to eliminate arrows, ALT byteacters, etc.*/ }
*name=0; return(j);}


int hasmates(char A)
{
int j=0;
for (j=0;j<25;j++)
   if ((j!=A)&&((RN[j]==RN[A])||(CN[j]==CN[A]))) return(1);
return(0);
}

int findreps(int reps[], char *text, int N)
/* finds repeat patterns in a piece of text of length N.
   stores the pattern in array reps[]. Returns the position
   of the first repeated digram, -1 if nothing found. */
{
int i,j,first=-1;
char A,B,C,D;
for (i=0;i<TIPMAX/2;i++) reps[i]=0;

for (i=0;i<N;i+=2)
   {
   A=*(text+i), B=*(text+i+1);
   for (j=i+2;j<N;j+=2)
      {
      C=*(text+j), D=*(text+j+1);
      if ((A==C)&&(B==D))     /* repeat found */
          {
          if (first==-1) first=i/2;
          reps[i/2]=(j-i)/2;
          break;
          }
      else if ((A==D)&&(B==C))     /* inversion found */
          {
          if (first==-1) first=i/2;
          reps[i/2]=-(j-i)/2;
          break;
          }
      }
   }
/* repeats has been constructed, first has been found */
return(first);
}